// Useful functions and such.

//#include "Multiforcer.h"

// file_size: Pass in path, get back file size in bytes.
long int file_size(const char *path) {
  struct stat file_status;
  
  if(stat(path, &file_status) != 0){
    printf("Unable to stat %s\n", path);
    return -1;
  }
  return file_status.st_size;
}

int convertAsciiToBinary(char *input, unsigned char *hash, int maxLength) {
  char convertSpace[10];
  unsigned int i;
  
  // Loop until either maxLength is hit, or strlen(intput) / 2 is hit.
  for (i = 0; (i < (unsigned int)maxLength) && (i < (strlen(input) / 2)); i++) {
    convertSpace[0] = input[2 * i];
    convertSpace[1] = input[2 * i + 1];
    convertSpace[2] = 0;
    sscanf(convertSpace, "%x", &hash[i]);
  }
  return i;
}

// Sort an array 
int sortHashes(struct HOST_Hash_Data* Sort_Hashes) {
  #define MAX_LEVELS 100000
  long int i, L, R, beg[MAX_LEVELS], end[MAX_LEVELS];
  unsigned char *piv;
  
  // Malloc piv
  piv = (unsigned char *) malloc(Sort_Hashes->hash_length * sizeof(unsigned char));

  // Quicksort pulled from http://alienryderflex.com/quicksort/
  i=0;
  beg[0]=0; 
  end[0]=Sort_Hashes->number_of_hashes;
  while (i>=0) {
    L=beg[i]; R=end[i]-1;
    if (L<R) {
      //piv=HashesToSort[L];
      memcpy(piv, &Sort_Hashes->hashes[L * Sort_Hashes->hash_length], Sort_Hashes->hash_length);
      
      // There's no way we will hit MAX_LEVELS
	  // Er... apparently there was.  Whoops.
      if (i==MAX_LEVELS-1) {printf("MAX LEVELS REACHED.  EXITING.  (use a smaller hash list or report the bug)\n"); exit(1);}
      while (L<R) {
        //while (greaterThanEqual(HashesToSort[R],piv) && L<R) R--; if (L<R) HashesToSort[L++]=HashesToSort[R];
        while ((memcmp(&Sort_Hashes->hashes[R * Sort_Hashes->hash_length], piv, Sort_Hashes->hash_length) >= 0) && L<R) R--;
        if (L<R) memcpy(&Sort_Hashes->hashes[L++ * Sort_Hashes->hash_length], &Sort_Hashes->hashes[R * Sort_Hashes->hash_length], Sort_Hashes->hash_length);
        //while (lessThanEqual(HashesToSort[L],piv) && L<R) L++; if (L<R) HashesToSort[R--]=HashesToSort[L]; 
        while ((memcmp(&Sort_Hashes->hashes[L * Sort_Hashes->hash_length], piv, Sort_Hashes->hash_length) <= 0) && L<R) L++;
        if (L<R) memcpy(&Sort_Hashes->hashes[R-- * Sort_Hashes->hash_length], &Sort_Hashes->hashes[L * Sort_Hashes->hash_length], Sort_Hashes->hash_length);                
        }
      //HashesToSort[L]=piv; beg[i+1]=L+1; end[i+1]=end[i]; end[i++]=L; 
      memcpy(&Sort_Hashes->hashes[L * Sort_Hashes->hash_length], piv, Sort_Hashes->hash_length);
      beg[i+1]=L+1; end[i+1]=end[i]; end[i++]=L; 
      }
    else {
      i--; 
    }
  }
  printf("Returning from sort.\n"); fflush(stdout);
  return true;
}

// Initialize the 8kb shared memory bitmap.
// 8192 bytes, 8 bits per byte.
// Bits are determined by the first 16 bits of a hash.
// Byte index is 13 bits, bit index is 3 bits.
// Algorithm: To set:
// First 13 bits of the hash (high order bits) are used as the
// index to the array.
// Next 3 bits control the left-shift amount of the '1' bit.
// So, hash 0x0000 has byte 0, LSB set.
// Hash 0x0105 has byte 
void initSharedBitmap(struct HOST_Hash_Data* Bitmap_Hashes) {
  uint32_t i;
  unsigned char bitmap_byte;
  uint32_t bitmap_index;
  // Zero bitmap
  memset(&Bitmap_Hashes->shared_bitmap, 0, 8192);
  
  for (i = 0; i < Bitmap_Hashes->number_of_hashes; i++) {
    // Load first two bytes of hash - reversed order to work with the little endian memory storage.  Otherwise the GPU has to flip things around
    // and this is operations that should not have to be performed.
    bitmap_index = (Bitmap_Hashes->hashes[i * Bitmap_Hashes->hash_length + 1] << 8) + Bitmap_Hashes->hashes[(i * Bitmap_Hashes->hash_length)];
    //printf("Hash %d: 2 bytes: %02X %02X ", i, Bitmap_Hashes->hashes[i * Bitmap_Hashes->hash_length], Bitmap_Hashes->hashes[i * Bitmap_Hashes->hash_length + 1]);
    //printf("Bitmap index: %04X\n", bitmap_index);
    // Shift left by the lowest 3 bits
    bitmap_byte = 0x01 << (bitmap_index & 0x0007);
    bitmap_index = bitmap_index >> 3;
    //printf(" Index %ld, byte %02X\n", bitmap_index, bitmap_byte);
    Bitmap_Hashes->shared_bitmap[bitmap_index] |= bitmap_byte;
  }
  if (false && verbose) {
    printf("Bitmap output\n");
    for (i = 0; i < 8192; i++) {
      if (i % 4 == 0) {
        printf("\n Index %d: ", i);
      }
      printf("%02X ", Bitmap_Hashes->shared_bitmap[i]);
    }
  }
}

// Does the same thing as above, but for a much larger bitmap for global memory space.
// Right now, this only works for 512MB hash space (32 bits worth of bitmap).
// Else returns a NULL pointer & does not do any work.

void initGlobalBitmap(struct HOST_Hash_Data* Bitmap_Hashes, uint32_t bits) {
  uint32_t i;
  unsigned char bitmap_byte;
  uint32_t bitmap_index;
  // Allocate bitmap
  // If the card does not have the RAM for it... too bad.
  if (cudaMalloc((void **)&Bitmap_Hashes->global_bitmap_device, 1 << (bits - 3)) == cudaErrorMemoryAllocation) {
    fprintf(stderr, "Unable to allocate %u MiB video card RAM.  Try without or smaller -l\n", 1 << (bits - 23));
    exit(1);
  } else {
    // Success!
    fflush(stdout);
    Bitmap_Hashes->global_bitmap_host = (unsigned char *)malloc(1 << (bits - 3));
    if (Bitmap_Hashes->global_bitmap_host == NULL) {
      fprintf(stderr, "Unable to allocate %u MiB host RAM - something has gone very wrong.  Try without or smaller -l\n", 1 << (bits - 23));
      exit(1);
    }
    memset(Bitmap_Hashes->global_bitmap_host, 0, 1 << (bits - 3));
  }

  
  for (i = 0; i < Bitmap_Hashes->number_of_hashes; i++) {
    // Load second 4 bytes of hash.  This is loaded normally, 
    bitmap_index = (((uint32_t)Bitmap_Hashes->hashes[i * Bitmap_Hashes->hash_length + 4]      ) | ((uint32_t)Bitmap_Hashes->hashes[(i * Bitmap_Hashes->hash_length + 5)] <<  8)
    		      | ((uint32_t)Bitmap_Hashes->hashes[i * Bitmap_Hashes->hash_length + 6] << 16) | ((uint32_t)Bitmap_Hashes->hashes[(i * Bitmap_Hashes->hash_length + 7)] << 24)) >> (32 - bits);
    //printf("Hash %ld: 4 bytes: %02X %02X %02X %02X", i, Bitmap_Hashes->hashes[i * Bitmap_Hashes->hash_length], 
    			//Bitmap_Hashes->hashes[i * Bitmap_Hashes->hash_length + 1], Bitmap_Hashes->hashes[i * Bitmap_Hashes->hash_length + 2],
    			//Bitmap_Hashes->hashes[i * Bitmap_Hashes->hash_length + 3]);
    //printf("Bitmap index: %08X\n", bitmap_index);
    // Shift left by the lowest 3 bits
    bitmap_byte = 0x01 << (bitmap_index & 0x00000007);
    bitmap_index = bitmap_index >> 3;
    //printf(" Index %ld, byte %02X\n", bitmap_index, bitmap_byte);
    Bitmap_Hashes->global_bitmap_host[bitmap_index] |= bitmap_byte;
  }
   cudaMemcpy(Bitmap_Hashes->global_bitmap_device, Bitmap_Hashes->global_bitmap_host, 1 << (bits - 3), cudaMemcpyHostToDevice);  
}


// Load unsalted passwords from a file.
// This requires a file of ascii-hex hashes, one per line.
// This does not handle salts.
struct HOST_Hash_Data* Load_Unsalted_Hashes(const char *hash_file_name, int hash_length) {
  FILE *hash_data_file;
  struct HOST_Hash_Data *Hash_Data;
  long int estimated_number_of_hashes;
  // If we have hashes with >4096 characters, there's a problem.  Though there's space for newlines.
  char buffer[4100];
  
  // Estimate the number of hashes as (file size / hash_length) - this will be higher than the actual value.
  estimated_number_of_hashes = file_size(hash_file_name) / hash_length;
  
  // Open hash file for reading, check for errors
  hash_data_file = fopen(hash_file_name, "r");
  if (!hash_data_file) {
    printf("Cannot load hash file %s.  Exiting.\n", hash_file_name);
    exit(1);
  }
  // Allocate memory and set up hash structure
  Hash_Data = (HOST_Hash_Data *) malloc(sizeof(HOST_Hash_Data));
  memset(Hash_Data, 0, sizeof(HOST_Hash_Data));  
  Hash_Data->hash_length = hash_length;
  Hash_Data->number_of_hashes = 0;
  Hash_Data->hashes = (unsigned char *) malloc(estimated_number_of_hashes * hash_length * sizeof(unsigned char));
  Hash_Data->max_salt_length = 0;
  Hash_Data->salt_length = NULL;
  Hash_Data->salt = NULL;

  // Load hashes into struct
  while (!feof(hash_data_file)) {
    // If fgets returns NULL, there's been an error or eof.  Continue.
    if (!fgets(buffer, 4100, hash_data_file)) {
      continue;
    }
    // If this is not a full line, continue (usually a trailing crlf)
    if (strlen(buffer) < ((unsigned int)hash_length * 2)) {
      continue;
    }

    convertAsciiToBinary(buffer, &Hash_Data->hashes[hash_length * Hash_Data->number_of_hashes], hash_length);
    Hash_Data->number_of_hashes++;
  }
  // Hashes have been loaded.  Sort by hash.
  sortHashes(Hash_Data);
  // Set up 8kb bitmap for hashes
  initSharedBitmap(Hash_Data);
  // Try to set up 512MB bitmap if desired
  if (big_lookup_table) {
    initGlobalBitmap(Hash_Data, big_lookup_table);
  }  
  // Initialize memory for other arrays
  Hash_Data->success = (unsigned char *)malloc(Hash_Data->number_of_hashes * sizeof(unsigned char));
  memset(Hash_Data->success, 0, Hash_Data->number_of_hashes * sizeof(unsigned char));
  Hash_Data->previous_success = (unsigned char *)malloc(Hash_Data->number_of_hashes * sizeof(unsigned char));
  memset(Hash_Data->previous_success, 0, Hash_Data->number_of_hashes * sizeof(unsigned char));  
  Hash_Data->passwords = (unsigned char *)malloc(Hash_Data->number_of_hashes * MAX_PASSWORD_LENGTH * sizeof(unsigned char));
  memset(Hash_Data->passwords, 0, Hash_Data->number_of_hashes * MAX_PASSWORD_LENGTH * sizeof(unsigned char));    
  return Hash_Data;
}
	



