/*  Multiforcer General Header File */

#include <stdlib.h>
#include "pstdint.h"
//#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>

typedef uint32_t UINT4;




// Structures for starting positions in character space
// This /should/ be two load commands, per docs (128 bits)
// Note that if MAX_CHARSET is >256, unsigned char may no longer be sufficient.
struct  __align__(16) start_positions {
  unsigned char p0;
  unsigned char p1;
  unsigned char p2;
  unsigned char p3;
  unsigned char p4;
  unsigned char p5;
  unsigned char p6;
  unsigned char p7;
  unsigned char p8;
  unsigned char p9;
  unsigned char p10;
  unsigned char p11;
  unsigned char p12;
  unsigned char p13;
  unsigned char p14;
  unsigned char p15;
};

// Contains host passwords/salts and appropriate data
//   hash_length: Length of each hash, in bytes
//   number_of_hashes: Number of loaded hashes/salts
//   *hashes: Pointer to array of hashes
//   max_salt_length: Maximum salt length present (array stride)
//   *salt_length: One byte per password, length of actual salt
//   *salt: Per-password, the salt
//   shared_bitmap: 8192 byte array for shared memory bitmap
//   *success: One byte per password, set to '1' if success.
//   *previous_success: One byte per password, for storing already-printed passwords
//   *passwords: Null-terminated passwords found
typedef struct HOST_Hash_Data {
  int hash_length;
  uint32_t number_of_hashes;
  unsigned char *hashes;
  int max_salt_length;
  unsigned char *salt_length;
  unsigned char *salt;
  unsigned char shared_bitmap[8192];
  unsigned char *global_bitmap_host;
  unsigned char *global_bitmap_device;
  unsigned char *success;
  unsigned char *previous_success;
  unsigned char *passwords;
}HOST_Hash_Data;

// Contains data related to a password test
//   CUDA_Device: Device to run on
//   password_length: Length of the password being tested
//   charset_length: Length of the character set
//   charset: Pointer to the charset
//   charset_lengths: Length of each element of the charset
//   target_ms: Target time for a kernel execution
//   *outputfile: File pointer to output data to, or NULL for no output.
//   blocks: Block count if set.
//   auto_blocks: Optimize block count
//   threads: Thread count if set.
//   auto_threads: Optimize thread count
typedef struct HOST_Run_Data {
  int CUDA_Device;
  int password_length;
  uint32_t charset_length;
  unsigned char *charset;
  unsigned char *charset_length_array;
  int target_ms;
  FILE *outputfile;
  uint32_t blocks;
  unsigned char auto_blocks;
  uint32_t threads;
  unsigned char auto_threads;
}HOST_Run_data;

// Functions

long int file_size(char *path);
