/* Cryptohaze.com Multiforcer: NTLM functions (requires MD4.cu) */

int CUDA_NTLM_Get_Thread_Count(int passlength) {
  int threads;
  switch (passlength) {
	case 4:
	case 5: 
	case 6: 
	  threads = 512;
	  break;
	case 7: 
	  threads = 448;
	  break;
	case 8: 
	case 9: 
	  threads = 384;
	  break;
  	case 10: 
	case 11: 
	case 12: 
	  threads = 320;
	  break;
	case 13: 
	case 14: 
	  threads = 256;
	  break;
  }
  return threads;
}





void CPU_NTLM_Search_0(unsigned char *SearchHash, unsigned char *Output, unsigned char *success, unsigned char *global_charset, int charsetLen, int numberOfPasswords) {

  unsigned char hash[16];
  unsigned char passwordString[64];
  int password;
  memset(passwordString, 0, 64);
  CPU_MD4(passwordString, 0, hash);
    // Password created.  Check against arrays.
    for (password = 0; password < numberOfPasswords; password++) {
      // If there is a match, deal with it appropriately.
      if (memcmp(hash, &SearchHash[HASH_LENGTH*password], HASH_LENGTH) == 0) {
        success[password] = 1;
      }
    }
}


#define NTLM_CUDA_KERNEL_CREATE(length) \
__global__ void CUDA_NTLM_Search_##length (unsigned char *OutputPassword, unsigned char *success,  \
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions,  \
				unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) { \
  const int pass_length = length; \
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15; \
  uint32_t a,b,c,d; \
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x; \
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes; \
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15; \
  UINT4 password_count = 0; \
  __shared__ __align__(16) unsigned char sharedCharset[4096]; \
  __shared__ __align__(16) unsigned char sharedBitmap[8192]; \
  __shared__ __align__(8) unsigned char sharedLengths[16]; \
  copyCharsetAndBitmap(sharedCharset, sharedBitmap, sharedLengths, charsetLen, pass_length); \
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions,  \
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15); \
  while (password_count < count) { \
  initNTLM(pass_length, sharedCharset, \
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, \
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	 \
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d); \
  checkHashMulti(pass_length, sharedBitmap, DEVICE_HashTable, numberOfPasswords, \
		DEVICE_Hashes_32, success, OutputPassword, \
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, \
		a, b, c, d, b0, b1, b2, b3, b4, b5); \
  password_count++; \
  incrementCounters##length##Multi(); \
  } \
}

#define FASTNTLM_CUDA_KERNEL_CREATE(length) \
__global__ void CUDA_FASTNTLM_Search_##length (unsigned char *OutputPassword, unsigned char *success,  \
			    int charsetLen, uint32_t numberOfPasswords, struct start_positions *DEVICE_Start_Positions,  \
				unsigned int count, unsigned char * DEVICE_Hashes, unsigned char *DEVICE_HashTable) { \
  const int pass_length = length; \
  uint32_t b0,b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15; \
  uint32_t a,b,c,d; \
  uint32_t thread_index = blockIdx.x*blockDim.x + threadIdx.x; \
  uint32_t *DEVICE_Hashes_32 = (uint32_t *)DEVICE_Hashes; \
  unsigned char p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15; \
  UINT4 password_count = 0; \
  __shared__ __align__(16) unsigned char sharedCharset[256]; \
  copyCharsetFAST(sharedCharset, charsetLen); \
  loadStartPositions(pass_length, thread_index, DEVICE_Start_Positions,  \
		   p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15); \
  while (password_count < count) { \
  initNTLMFAST(pass_length, sharedCharset, \
  	p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, \
	b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15);	 \
  CUDA_MD4(b0, b1, b2, b3, b4, b5, b6, b7, b8, b9, b10, b11, b12, b13, b14, b15, a, b, c, d); \
  checkHashFAST(pass_length, numberOfPasswords, \
		DEVICE_Hashes_32, success, OutputPassword, \
		p0, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15, \
		a, b, c, d, b0); \
  password_count++; \
  incrementCounters##length (); \
  } \
}

NTLM_CUDA_KERNEL_CREATE(1)
NTLM_CUDA_KERNEL_CREATE(2)
NTLM_CUDA_KERNEL_CREATE(3)
NTLM_CUDA_KERNEL_CREATE(4)
NTLM_CUDA_KERNEL_CREATE(5)
NTLM_CUDA_KERNEL_CREATE(6)
NTLM_CUDA_KERNEL_CREATE(7)
NTLM_CUDA_KERNEL_CREATE(8)
NTLM_CUDA_KERNEL_CREATE(9)
NTLM_CUDA_KERNEL_CREATE(10)
NTLM_CUDA_KERNEL_CREATE(11)
NTLM_CUDA_KERNEL_CREATE(12)
NTLM_CUDA_KERNEL_CREATE(13)
NTLM_CUDA_KERNEL_CREATE(14)
NTLM_CUDA_KERNEL_CREATE(15)
NTLM_CUDA_KERNEL_CREATE(16)

FASTNTLM_CUDA_KERNEL_CREATE(1)
FASTNTLM_CUDA_KERNEL_CREATE(2)
FASTNTLM_CUDA_KERNEL_CREATE(3)
FASTNTLM_CUDA_KERNEL_CREATE(4)
FASTNTLM_CUDA_KERNEL_CREATE(5)
FASTNTLM_CUDA_KERNEL_CREATE(6)
FASTNTLM_CUDA_KERNEL_CREATE(7)
FASTNTLM_CUDA_KERNEL_CREATE(8)
FASTNTLM_CUDA_KERNEL_CREATE(9)
FASTNTLM_CUDA_KERNEL_CREATE(10)
FASTNTLM_CUDA_KERNEL_CREATE(11)
FASTNTLM_CUDA_KERNEL_CREATE(12)
FASTNTLM_CUDA_KERNEL_CREATE(13)
FASTNTLM_CUDA_KERNEL_CREATE(14)
FASTNTLM_CUDA_KERNEL_CREATE(15)
FASTNTLM_CUDA_KERNEL_CREATE(16)

void Launch_CPU_NTLM_Kernel(int passlength, unsigned char *SearchHash, unsigned char *Output, unsigned char *success, 
							unsigned char *global_charset, int charsetLen, int numberOfPasswords) {
  if (passlength == 0) {
    CPU_NTLM_Search_0(SearchHash, Output, success, global_charset, charsetLen, numberOfPasswords);
  }
}


void Launch_CUDA_NTLM_Kernel(int passlength, uint32_t charsetLength, int numberOfPasswords, unsigned char *DEVICE_Passwords, 
						unsigned char *DEVICE_Success, struct start_positions *DEVICE_Start_Positions, uint32_t per_step, uint32_t threads, uint32_t blocks, unsigned char *DEVICE_Hashes, unsigned char *DEVICE_Bitmap) {
    if (passlength == 1) {
	  CUDA_NTLM_Search_1 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 2) {
	  CUDA_NTLM_Search_2 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 3) {
	  CUDA_NTLM_Search_3 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 4) {
	  CUDA_NTLM_Search_4 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 5) {
	  CUDA_NTLM_Search_5 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 6) {
	  CUDA_NTLM_Search_6 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 7) {
	  CUDA_NTLM_Search_7 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 8) {
	  CUDA_NTLM_Search_8 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);  
	} else if (passlength == 9) {
	  CUDA_NTLM_Search_9 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 10) {
	  CUDA_NTLM_Search_10 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 11) {
	  CUDA_NTLM_Search_11 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 12) {
	  CUDA_NTLM_Search_12 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 13) {
	  CUDA_NTLM_Search_13 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 14) {
	  CUDA_NTLM_Search_14 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 15) {
	  CUDA_NTLM_Search_15 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 16) {
	  CUDA_NTLM_Search_16 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	}
	
	cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
      {
        fprintf(stderr, "Cuda error: %s.\n", cudaGetErrorString( err) );
        exit(EXIT_FAILURE);
      } 
}

void Launch_CUDA_FASTNTLM_Kernel(int passlength, uint32_t charsetLength, int numberOfPasswords, unsigned char *DEVICE_Passwords, 
						unsigned char *DEVICE_Success, struct start_positions *DEVICE_Start_Positions, uint32_t per_step, uint32_t threads, uint32_t blocks, unsigned char *DEVICE_Hashes, unsigned char *DEVICE_Bitmap) {
    if (passlength == 1) {
	  CUDA_FASTNTLM_Search_1 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 2) {
	  CUDA_FASTNTLM_Search_2 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 3) {
	  CUDA_FASTNTLM_Search_3 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 4) {
	  CUDA_FASTNTLM_Search_4 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 5) {
	  CUDA_FASTNTLM_Search_5 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 6) {
	  CUDA_FASTNTLM_Search_6 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 7) {
	  CUDA_FASTNTLM_Search_7 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 8) {
	  CUDA_FASTNTLM_Search_8 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);  
	} else if (passlength == 9) {
	  CUDA_FASTNTLM_Search_9 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 10) {
	  CUDA_FASTNTLM_Search_10 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 11) {
	  CUDA_FASTNTLM_Search_11 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 12) {
	  CUDA_FASTNTLM_Search_12 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 13) {
	  CUDA_FASTNTLM_Search_13 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap); 
	} else if (passlength == 14) {
	  CUDA_FASTNTLM_Search_14 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 15) {
	  CUDA_FASTNTLM_Search_15 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	} else if (passlength == 16) {
	  CUDA_FASTNTLM_Search_16 <<< blocks, threads >>> (DEVICE_Passwords, DEVICE_Success, charsetLength, numberOfPasswords, DEVICE_Start_Positions, per_step, DEVICE_Hashes, DEVICE_Bitmap);
	}
	
	cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
      {
        fprintf(stderr, "Cuda error: %s.\n", cudaGetErrorString( err) );
        exit(EXIT_FAILURE);
      } 
}


